/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.util;

import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.Window;
import java.util.Properties;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import ca.uvic.csr.shrimp.DataBean.DataBean;
import ca.uvic.csr.shrimp.DisplayBean.DisplayBean;
import ca.uvic.csr.shrimp.ShrimpApplication.ApplicationAccessor;

/**
 * @author Rob Lintern
 *
 * A collection of general utilites for SHriMP
 */
public class ShrimpUtils {

	/**
	 * Compares the two objects and returns true if they are equal.<br>
	 * If both are null then <b>true</b> is retured.<br>
	 * If one object is null and the other is not null then <b>false</b> is returned.<br>
	 * Otherwise if both objects are not null then the .equals method is used to compared the two.
	 * @param oldValue
	 * @param newValue
	 * @return boolean
	 */
	public static boolean equals(Object oldValue, Object newValue) {
		boolean equal;	
		// case 1: both are null 
		if ((oldValue == null) && (newValue == null)) {
			equal = true;
		} 
		// case 2: neither are null
		else if ((oldValue != null) && (newValue != null)) {
			equal = oldValue.equals(newValue);
		} 
		// case 3: one is null, the other is not null 
		else {
			equal = false;
		}
		return equal;
	}
	
	/**
	 * Determines if the bsf.jar and js.jar libraries are in the classpath.
	 * These are needed for Scripting.
	 * @return true if scripting libraries are installed
	 */
	public static boolean isScriptingToolInstalled() {
		try {
		    new org.apache.bsf.BSFException("Test");
		    new org.mozilla.javascript.Label();
		    return true;
		} catch (Throwable ignore) {
			return false;
		}
	}
	
    public static boolean shouldLoadAllData() {
	    boolean shouldLoadAll = true;
		Component parentComp = ApplicationAccessor.getApplication().getParentFrame();
		String title = ApplicationAccessor.getApplication().getName();
		String message = "This action will cause all data to be loaded into memory which, for large projects, " +
							"\ncould take a long time and use up a lot of memory.";			
		message += "Are you sure you want to proceed?";
		int result = JOptionPane.showConfirmDialog(parentComp, message, title, JOptionPane.YES_NO_OPTION);
		shouldLoadAll = result == JOptionPane.OK_OPTION;
	    return shouldLoadAll;       
    }
    
	public static boolean shouldChangeCprels(String [] cprels, boolean inverted, String [] defaultCprels, boolean defaultCprelsInverted) {
	    // check if user wants to load all data, if not using the default hierarchy
		// **NOTE: there is an assumption made here that choosing the default hierarchy doesn't 
		// require all data to be loaded from the back-end. So far (Sept 30, 2004), this is the case but
		// may not always be so for any new domains.
	    if (!CollectionUtils.haveSameElements(cprels, defaultCprels) || inverted != defaultCprelsInverted) {
	        return shouldLoadAllData();
	    }
	    return true;
	}
	

	public static boolean shouldShowManyRoots(String [] cprels, boolean inverted, DataBean dataBean) {
	    Vector artifacts = cprels.length > 0 && inverted ? dataBean.getLeafArtifacts(cprels) : dataBean.getRootArtifacts(cprels);
	    return shouldShowManyRoots(artifacts.size());
	}
	
	public static boolean shouldShowManyRoots(int numRoots) {
	    boolean shouldShow = true;
	    // check if too many roots to show
		Properties properties = ApplicationAccessor.getApplication().getProperties();
		String showWarningStr = properties.getProperty(DisplayBean.PROPERTY_KEY__SHOW_MANY_CHILDREN_WARNING, DisplayBean.DEFAULT_SHOW_MANY_CHILDREN_WARNING);
		boolean showWarning = (new Boolean(showWarningStr)).booleanValue();
		if (showWarning) {
			String thresholdStr = properties.getProperty(DisplayBean.PROPERTY_KEY__SHOW_MANY_CHILDREN_WARNING_THRESHOLD, DisplayBean.DEFAULT_SHOW_MANY_CHILDREN_WARNING_THRESHOLD);
			int threshold = (new Integer(thresholdStr)).intValue();
			if (numRoots > threshold) {
				Component parentComp = ApplicationAccessor.getApplication().getParentFrame();
				String title = ApplicationAccessor.getApplication().getName();
				String message = "";
				message += "All data has been loaded from the back-end, but there will be " + numRoots + " nodes!" +
							"\nIt may take some time to create them all.";
				message += "\n\nAre you sure you want to continue?";
				int result = JOptionPane.showConfirmDialog(parentComp, message, title, JOptionPane.YES_NO_OPTION);
				shouldShow = (result == JOptionPane.OK_OPTION);
			} else {
				shouldShow = true;
			}
		}
		return shouldShow;
	}
	
	public static boolean hasRootArtifacts(String [] cprels, boolean inverted, DataBean dataBean) {
	    Vector artifacts = cprels.length > 0 && inverted ? dataBean.getLeafArtifacts(cprels) : dataBean.getRootArtifacts(cprels);
	    if (artifacts.isEmpty()) {
	        // dont bother changing the hierarchy if there are no root artifacts.
			Object message = "The hierachy of \"" + cprelsToString(cprels) + "\" has no root nodes!" +
			"\nThe hierarchy will not be changed.";					
	        JOptionPane.showMessageDialog(ApplicationAccessor.getApplication().getParentFrame(), message, ApplicationAccessor.getApplication().getName(), JOptionPane.INFORMATION_MESSAGE);
	    }
	    return !artifacts.isEmpty();
	}
	
	public static String cprelsToString(String [] cprels) {
		String s = "";
		for (int i = 0; i < cprels.length; i++) {
			String cprel = cprels[i];
			if (i > 0) {
				s += ", ";
			}
			s += cprel;
		}
		return s;
	}
	
	public static int showDialogWithPropertyCheckbox(Object message, int optionType) {	    
		Component parentComp = ApplicationAccessor.getApplication().getParentFrame();
		String title = ApplicationAccessor.getApplication().getName();
		int result = JOptionPane.showConfirmDialog(parentComp, message, title, optionType);
		return result;
	}

	public static Dialog getParentDialog(Component parent) {
		Window window = SwingUtilities.windowForComponent(parent);
		return (window instanceof Dialog ? (Dialog) window : null);
	}
	
	public static Frame getParentFrame(Component parent) {
		Window window = SwingUtilities.windowForComponent(parent);
		return (window instanceof Frame ? (Frame) window : null);
	}
	
	/**
	 * Centers the given window on the window for the parent component.
	 * Doesn't allow the window to be placed off screen.
	 */
	public static void centerWindowOnParent(Window window, Component parent) {
		Window parentWindow = null;
		if (parent != null) {
			parentWindow = (parent instanceof Window ? (Window) parent : SwingUtilities.windowForComponent(parent));
		}
		
		int x = 0, y = 0;
		if (parentWindow != null) {
			x = parentWindow.getX() + (parentWindow.getWidth() / 2) - (window.getWidth() / 2);
			y = parentWindow.getY() + (parentWindow.getHeight() / 2) - (window.getHeight() / 2);

			// ensure that the window won't be off the edge off the bottom or right side of the screen
			Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
			if ((x + window.getWidth()) > screenSize.width) {
				x = screenSize.width - window.getWidth();
			}
			if ((y + window.getHeight()) > screenSize.height) {
				y = screenSize.height - window.getWidth();
			}
			// ensure that the window is not off the top or left side of the screen
			x = Math.max(0, x);
			y = Math.max(0, y);
		}

		window.setLocation(x, y);
	}

	
}
